import base64


def base64_string(val):
    if val is None:
        return val
    return base64.b64encode(val).decode('utf-8')


def to_int(val):
    if val is None:
        return val
    if val == b'':
        return 0
    if isinstance(val, bytes):
        return int.from_bytes(val, 'big')
    if isinstance(val, str):
        return int(val)

    return val



# MIT License
#
# Copyright (c) 2020 Evgeny Medvedev, evge.medvedev@gmail.com
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from iotexetl.jobs import RETRY_EXCEPTIONS
from iotexetl.mappers.evm_transfer_mapper import map_evm_transfers
from iotexetl.service.iotex_service import IotexService
from blockchainetl_common.executors.batch_work_executor import BatchWorkExecutor
from blockchainetl_common.jobs.base_job import BaseJob
from blockchainetl_common.utils import validate_range


# Exports blocks, actions, and receipts
class ExportEvmTransfersJob(BaseJob):
    def __init__(
            self,
            start_block,
            end_block,
            iotex_rpc,
            max_workers,
            item_exporter):
        validate_range(start_block, end_block)
        self.start_block = start_block
        self.end_block = end_block

        self.batch_work_executor = BatchWorkExecutor(1, max_workers, retry_exceptions=RETRY_EXCEPTIONS)
        self.item_exporter = item_exporter

        self.iotex_service = IotexService(iotex_rpc)

    def _start(self):
        self.item_exporter.open()

    def _export(self):
        self.batch_work_executor.execute(
            range(self.start_block, self.end_block + 1),
            self._export_batch,
            total_items=self.end_block - self.start_block + 1
        )

    def _export_batch(self, block_number_batch):
        assert len(block_number_batch) == 1
        block_number = block_number_batch[0]
        block_evm_transfers = self.iotex_service.get_evm_transfers(block_number)
        if block_evm_transfers:
            block = self.iotex_service.get_block(block_evm_transfers.blockHeight)
            self.item_exporter.export_items(map_evm_transfers(block, block_evm_transfers))

    def _end(self):
        self.batch_work_executor.shutdown()
        self.item_exporter.close()



from iotexetl.utils import string_utils


def map_evm_transfers(raw_block, block_evm_transfers):
    for action in block_evm_transfers.actionEvmTransfers:
        for transfer in action.evmTransfers:
            yield {
                'type': 'evm_transfer',
                'height': block_evm_transfers.blockHeight,
                'action_hash': action.actionHash.hex(),
                'amount': string_utils.to_int(transfer.amount),
                'from': getattr(transfer, 'from'),
                'to': transfer.to,
                'timestamp': raw_block.block.header.core.timestamp.ToJsonString(),
            }


# MIT License
#
# Copyright (c) 2020 Evgeny Medvedev, evge.medvedev@gmail.com
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from iotexetl.utils.string_utils import base64_string

def map_block(raw_block, raw_block_meta):
    header = raw_block.block.header
    block = {
        'type': 'block',
        'version': header.core.version,
        'height': header.core.height,
        'hash': raw_block_meta.hash,
        'timestamp': header.core.timestamp.ToJsonString(),
        'prev_block_hash': header.core.prevBlockHash.hex(),
        'tx_root': base64_string(header.core.txRoot),
        'receipt_root': base64_string(header.core.receiptRoot),
        'delta_state_digest': base64_string(header.core.deltaStateDigest),
        'producer_pubkey': base64_string(header.producerPubkey),
        'signature': base64_string(header.signature),
        'number_of_actions': len(raw_block.block.body.actions)
    }

    return block


from grpc import RpcError
from requests.exceptions import Timeout as RequestsTimeout, HTTPError, TooManyRedirects
from blockchainetl_common.executors.retriable_value_error import RetriableValueError

RETRY_EXCEPTIONS = (
    ConnectionError, HTTPError, RequestsTimeout, TooManyRedirects, OSError, RetriableValueError,
    RpcError)



from iotexetl.utils import string_utils


def map_implicit_transfer_log(raw):
    for action in raw.implicitTransferLog:
        for transaction in action.transactions:
            yield {
                'type': 'implicit_transfer_log',
                'action_hash': action.actionHash.hex(),
                'topic': string_utils.base64_string(transaction.topic),
                'amount': transaction.amount,
                'sender': transaction.sender,
                'recipient': transaction.recipient,
            }


import bech32
from eth_hash.auto import keccak as keccak_256


def pubkey_to_address(pubkey, prefix='io'):
    """This implements the algorithm described here https://github.com/iotexproject/iotex-address"""
    if pubkey is None or len(pubkey) < 1:
        return None
    pubkey_hash = keccak_256(pubkey[1:])
    if pubkey_hash is None or len(pubkey_hash) < 12:
        return None
    payload = pubkey_hash[12:]
    return bech32_encode(prefix, payload)


def pubkey_hex_to_address(pubkey_hex):
    if pubkey_hex is None:
        return None
    return pubkey_to_address(bytearray.fromhex(pubkey_hex))


def bech32_encode(hrp, witprog):
    five_bit_witprog = bech32.convertbits(witprog, 8, 5)
    if five_bit_witprog is None:
        return None
    ret = bech32.bech32_encode(hrp, five_bit_witprog)
    return ret
